from unittest.mock import PropertyMock

import pytest
from github.GithubException import GithubException

from ghsearch.filters import ContentFilter, FilterException, NotArchivedFilter, PathFilter, RegexContentFilter

from . import build_mock_content_file


@pytest.mark.parametrize(
    "path_matcher, path, expected_result",
    [
        ("file.py", "path/to/file.py", True),
        ("path", "path/to/file.py", True),
        ("to/file", "path/to/file.py", True),
        ("other.py", "path/to/file.py", False),
    ],
)
def test_build_path_filter(path_matcher, path, expected_result):
    path_filter = PathFilter(path_matcher)
    mock_content_file = build_mock_content_file(path=path)

    assert path_filter(mock_content_file) is expected_result
    assert path_filter.uses_core_api is False


@pytest.mark.parametrize(
    "content_matcher, content_bytes, expected_result",
    [
        ("this str", b"I'm looking for this string.", True),
        ("another str", b"I'm still looking for this str", False),
    ],
)
def test_build_content_filter(content_matcher, content_bytes, expected_result):
    content_filter = ContentFilter(content_matcher)
    mock_content_file = build_mock_content_file(decoded_content=content_bytes)

    assert content_filter(mock_content_file) is expected_result
    assert content_filter.uses_core_api is True


@pytest.mark.parametrize(
    "content_matcher, content_bytes, expected_result",
    [
        ("regex\\s{1}test", b"I'm looking for a regex test str", True),
        ("regex\\s{2}test", b"I'm looking for a regex test str", False),
    ],
)
def test_build_regex_content_filter(content_matcher, content_bytes, expected_result):
    content_filter = RegexContentFilter(content_matcher)
    mock_content_file = build_mock_content_file(decoded_content=content_bytes)

    assert content_filter(mock_content_file) is expected_result
    assert content_filter.uses_core_api is True


def test_build_regex_content_filter_invalid_regex():
    with pytest.raises(
        FilterException,
        match="Failed to compile regular expression from '\\[invalid regex': unterminated character set at position 0",
    ):
        RegexContentFilter("[invalid regex")


def test_content_filter_with_github_exception():
    content_filter = ContentFilter("something")
    mock_content_file = build_mock_content_file()
    type(mock_content_file).decoded_content = PropertyMock(side_effect=GithubException(403, {"message": "fail"}))

    with pytest.raises(FilterException, match="Error reading content from org/repo/path: fail") as exc_info:
        content_filter(mock_content_file)

    assert exc_info.value.filter == content_filter


@pytest.mark.parametrize(
    "archived, expected_result",
    [
        (False, True),
        (True, False),
    ],
)
def test_build_not_archived_filter(archived, expected_result):
    not_archived_filter = NotArchivedFilter()
    mock_content_file = build_mock_content_file(archived=archived)

    assert not_archived_filter(mock_content_file) is expected_result
    assert not_archived_filter.uses_core_api is True


def test_not_archived_filter_caches_access_to_archived_property():
    not_archived_filter = NotArchivedFilter()

    archived1_1 = PropertyMock(return_value=True)
    repo1_1 = build_mock_content_file("org/repo1", "file1.txt")
    type(repo1_1.repository).archived = archived1_1

    archived1_2 = PropertyMock(return_value=True)
    repo1_2 = build_mock_content_file("org/repo1", "file2.txt")
    type(repo1_2.repository).archived = archived1_2

    archived2_1 = PropertyMock(return_value=True)
    repo2_1 = build_mock_content_file("org/repo2", "file1.txt")
    type(repo2_1.repository).archived = archived2_1

    not_archived_filter(repo1_1)
    not_archived_filter(repo1_2)
    not_archived_filter(repo2_1)

    archived1_1.assert_called_once()
    archived1_2.assert_not_called()
    archived2_1.assert_called_once()



from unittest.mock import call, patch

import pytest

from ghsearch.terminal import ProgressPrinter


@pytest.mark.parametrize(
    "overwrite, isatty, force, expected_calls",
    [
        # overwrite = True, isatty = True, force = False
        (
            True,
            True,
            False,
            [
                call("\r\033[?25lHello!", nl=False),
                call("\r\033[?25l      ", nl=False),
                call("\r\033[?25h", nl=False),
            ],
        ),
        # overwrite = True, isatty = True, force = True
        (
            True,
            True,
            True,
            [
                call("\nHello!"),
                call("\r\033[?25l", nl=False),
                call("\r\033[?25h", nl=False),
            ],
        ),
        # overwrite = False, isatty = True, force = False
        (False, True, False, [call("Hello!")]),
        # overwrite = False, isatty = True, force = True
        (False, True, True, [call("Hello!")]),
        # overwrite = True, isatty = False, force = False
        (True, False, False, []),
        # overwrite = True, isatty = False, force = True
        (True, False, True, []),
        # overwrite = False, isatty = False, force = False
        (False, False, False, [call("Hello!")]),
        # overwrite = False, isatty = False, force = True
        (False, False, True, [call("Hello!")]),
    ],
)
def test_progress_printer(assert_click_echo_calls, overwrite, isatty, force, expected_calls):
    with patch("ghsearch.terminal.sys") as mock_sys:
        mock_sys.stdout.isatty.return_value = isatty

        with ProgressPrinter(overwrite=overwrite) as printer:
            printer("Hello!", force=force)

    assert_click_echo_calls(*expected_calls)



import contextlib
import re
import sys

import click

HIDE_CURSOR = "\033[?25l"
SHOW_CURSOR = "\033[?25h"
_ansi_re = re.compile(r"\033\[[;?0-9]*[a-zA-Z]")


def _strip_ansi(value):
    return _ansi_re.sub("", value)


def _term_len(x):
    return len(_strip_ansi(x))


class ProgressPrinter(contextlib.AbstractContextManager):
    def __init__(self, overwrite=True):
        self.overwrite = overwrite and sys.stdout.isatty()
        self.verbose = not overwrite
        self.last_width = 0

    def __enter__(self):
        def printer(message, force=False):
            if self.overwrite:
                if force:
                    click.echo(f"\n{message}")
                else:
                    self._overwrite_previous_line(message)
            elif self.verbose:
                click.echo(message)

        return printer

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.overwrite:
            self._overwrite_previous_line()
            click.echo("\r" + SHOW_CURSOR, nl=False)
        return None

    def _overwrite_previous_line(self, message=""):
        line_width = _term_len(message)
        click.echo(f"\r{HIDE_CURSOR}{message}{' ' * (self.last_width - line_width)}", nl=False)
        self.last_width = line_width



from typing import List

import click
import github
from github.ContentFile import ContentFile
from github.GithubException import GithubException
from github.Rate import Rate
from github.RateLimit import RateLimit

from ghsearch.filters import Filter, FilterException
from ghsearch.terminal import ProgressPrinter

CORE_CALLS_RELATIVE_LIMIT = 0.1
CORE_CALLS_ABSOLUTE_LIMIT = 500


def _confirm_continue_many_calls(core_rate: Rate, num_results: int, calls_per_res: int) -> None:
    click.confirm(
        f"""
Warning: you are about to potentially make more than {CORE_CALLS_ABSOLUTE_LIMIT} core api requests.
Your search yielded {num_results} results, and gh-search may make up to {calls_per_res} core api call(s) per result.

Your current core api usage is {core_rate.remaining}/{core_rate.limit} (resets {core_rate.reset})

Do you want to continue?""".strip(),
        abort=True,
    )


def _confirm_continue_near_limit(core_rate: Rate, num_results: int, calls_per_res: int) -> None:
    click.confirm(
        f"""
Warning: you are at risk of going below {CORE_CALLS_RELATIVE_LIMIT:.0%} of your remaining core api rate limit.
Your search yielded {num_results} results, and gh-search may make up to {calls_per_res} core api call(s) per result.

Your current core api usage is {core_rate.remaining}/{core_rate.limit} (resets {core_rate.reset})

Do you want to continue?""".strip(),
        abort=True,
    )


def _echo_rate_limits(rate_limit: RateLimit) -> None:
    click.echo(
        f"Core rate limit: {rate_limit.core.remaining}/{rate_limit.core.limit} (resets {rate_limit.core.reset}), "
        f"Search rate limit: {rate_limit.search.remaining}/{rate_limit.search.limit} (resets {rate_limit.search.reset})"
    )


class GHSearch:
    def __init__(self, client: github.Github, filters: List[Filter], verbose: bool = False):
        self.client = client
        self.filters = filters
        self.verbose = verbose

    def get_rate_limit(self) -> RateLimit | None:
        try:
            return self.client.get_rate_limit()
        except GithubException as ge:
            # 404 means that rate limiting is disabled
            if ge.status == 404:
                return None
            raise ge

    def get_filtered_results(self, query: List[str]) -> List[ContentFile]:
        rate_limit = self.get_rate_limit()

        if rate_limit and self.verbose:
            _echo_rate_limits(rate_limit)

        results = self.client.search_code(query=" ".join(query))

        if rate_limit:
            self._check_core_limit_threshold(results.totalCount, rate_limit.core)

        filtered_results = []
        with ProgressPrinter(overwrite=not self.verbose) as printer:
            for result in results:
                printer(f"Checking result for {result.repository.full_name}")
                try:
                    exclude_reason = self._should_exclude(result)
                except FilterException as e:
                    printer(str(e), force=True)
                else:
                    if not exclude_reason:
                        filtered_results.append(result)
                    elif self.verbose:
                        click.echo(f"Skipping result for {result.repository.full_name} via {exclude_reason}")

        rate_limit = self.get_rate_limit()
        if rate_limit and self.verbose:
            _echo_rate_limits(rate_limit)

        return filtered_results

    def _should_exclude(self, result):
        for result_filter in self.filters:
            if not result_filter(result):
                return result_filter.__class__.__name__
        return False

    def _check_core_limit_threshold(self, num_results: int, core_rate: Rate) -> None:
        max_core_api_calls_per_result = sum(bool(f.uses_core_api) for f in self.filters)
        if max_core_api_calls_per_result > 0:

            num_core_api_calls_worst_case = num_results * max_core_api_calls_per_result
            remaining_worst_case = core_rate.remaining - num_core_api_calls_worst_case
            if remaining_worst_case / core_rate.limit < CORE_CALLS_RELATIVE_LIMIT:
                _confirm_continue_near_limit(core_rate, num_results, max_core_api_calls_per_result)
            elif num_core_api_calls_worst_case > CORE_CALLS_ABSOLUTE_LIMIT:
                _confirm_continue_many_calls(core_rate, num_results, max_core_api_calls_per_result)



from collections import defaultdict
from typing import IO, Dict, List, Type
from urllib import parse

from github.ContentFile import ContentFile


class Printer:
    @staticmethod
    def sanitize_qualifiers_for_search_url(query: List[str]) -> List[str]:
        return [q for q in query if not (q.startswith("repo:") or q.startswith("org:"))]

    def __init__(self, stream: IO) -> None:
        self._stream = stream

    def print(self, query: List[str], results: List[ContentFile]) -> None:
        results_per_repo = defaultdict(list)
        for result in results:
            results_per_repo[result.repository.full_name].append(result)
        self._print(query, results_per_repo)
        self._stream.flush()

    def _print(self, query: List[str], results_per_repo: Dict[str, List[ContentFile]]) -> None:
        raise NotImplementedError()


_REGISTRY: Dict[str, Type[Printer]] = {}


def register_printer(cls: Type[Printer]) -> Type[Printer]:
    name = getattr(cls, "NAME", False)
    if not isinstance(name, str):
        raise NotImplementedError(f"{cls.__name__} must declare a NAME property")
    if name in _REGISTRY:
        raise IndexError(f"Printer {name} is already registered")
    _REGISTRY[name] = cls
    return cls


def printers_list() -> List[str]:
    return list(_REGISTRY.keys())


def printer_factory(name: str, stream: IO, force_repo_list_printer: bool = False) -> Printer:
    if force_repo_list_printer:
        return RepoListPrinter(stream)
    return _REGISTRY[name](stream)


@register_printer
class DefaultPrinter(Printer):
    NAME = "default"

    def _print(self, query: List[str], results_per_repo: Dict[str, List[ContentFile]]) -> None:
        if len(results_per_repo) == 0:
            self._stream.write("No results!\n")
            self._stream.write(
                "(For limitations of GitHub's code search see https://docs.github.com/en/github/"
                "searching-for-information-on-github/searching-code#considerations-for-code-search)\n"
            )
            return

        sorted_results = sorted(results_per_repo.items(), key=lambda kv: len(kv[1]), reverse=True)

        q_param = parse.quote(" ".join(self.sanitize_qualifiers_for_search_url(query)))
        self._stream.write("Results:\n")
        for repo, repo_results in sorted_results:
            repo_result = repo_results[0]
            url = f"{repo_result.repository.html_url}/search?utf8=✓&q={q_param}"
            self._stream.write(f" {len(repo_results)} - {repo}: {url}\n")

            repo_results.sort(key=lambda x: x.path)
            for result in repo_results:
                self._stream.write(f"\t- {result.path}\n")


@register_printer
class RepoListPrinter(Printer):
    NAME = "repo-list"

    def _print(self, query: List[str], results_per_repo: Dict[str, List[ContentFile]]) -> None:
        for repo in results_per_repo:
            self._stream.write(repo + "\n")


class StructuredPrinter(Printer):
    def _print(self, query: List[str], results_per_repo: Dict[str, List[ContentFile]]) -> None:
        self._print_serialise([self._build_repo_results(results_per_repo[repo]) for repo in results_per_repo])

    @classmethod
    def _build_repo_results(cls, results: List[ContentFile]) -> Dict:
        repo = results[0].repository
        return {
            "full_name": repo.full_name,
            "html_url": repo.html_url,
            "fork": repo.fork,
            "owner": repo.owner.login,
            "name": repo.name,
            "results": cls._build_results(results),
        }

    @classmethod
    def _build_results(cls, results: List[ContentFile]) -> List[Dict]:
        return [cls._build_result(result) for result in results]

    @staticmethod
    def _build_result(result: ContentFile) -> Dict:
        return {
            "path": result.path,
            "name": result.name,
            "size": result.size,
            "html_url": result.html_url,
        }

    def _print_serialise(self, structured_results: List[Dict]) -> None:
        raise NotImplementedError()


@register_printer
class JsonPrinter(StructuredPrinter):
    NAME = "json"

    def _print_serialise(self, structured_results: List[Dict]) -> None:
        import json

        json.dump(structured_results, fp=self._stream)


@register_printer
class YamlPrinter(StructuredPrinter):
    NAME = "yaml"

    def _print_serialise(self, structured_results: List[Dict]) -> None:
        from ruamel.yaml import YAML

        yaml = YAML(typ="safe", pure=True)
        yaml.default_flow_style = False
        yaml.dump(structured_results, stream=self._stream)



from unittest.mock import patch

import pytest


@pytest.fixture
def mock_click_echo():
    with patch("click.echo") as mock:
        yield mock


@pytest.fixture
def assert_click_echo_calls(mock_click_echo):
    def _fn(*calls):
        mock_click_echo.assert_has_calls(calls, any_order=False)
        assert mock_click_echo.call_count == len(calls)

    return _fn



from io import StringIO

import pytest

from ghsearch.output import DefaultPrinter, JsonPrinter, RepoListPrinter, YamlPrinter

from . import build_mock_content_file


@pytest.mark.parametrize(
    "printer_cls, expected",
    [
        (
            DefaultPrinter,
            "No results!\n"
            "(For limitations of GitHub's code search see https://docs.github.com/en/github/"
            "searching-for-information-on-github/searching-code#considerations-for-code-search)\n",
        ),
        (RepoListPrinter, ""),
    ],
)
def test_print_no_results(printer_cls, expected):
    stream = StringIO()
    printer_cls(stream).print(["query"], {})
    assert stream.getvalue() == expected


@pytest.mark.parametrize(
    "printer_cls, expected",
    [
        (
            DefaultPrinter,
            "Results:\n"
            " 2 - org/repo1: https://www.github.com/org/repo1/search?utf8=✓&q=query\n"
            "\t- README.md\n"
            "\t- file.txt\n",
        ),
        (RepoListPrinter, "org/repo1\n"),
    ],
)
def test_print_results_slingle_repo(printer_cls, expected):
    stream = StringIO()
    printer_cls(stream).print(
        ["query"],
        [
            build_mock_content_file("org/repo1", "README.md"),
            build_mock_content_file("org/repo1", "file.txt"),
        ],
    )

    assert stream.getvalue() == expected


@pytest.mark.parametrize(
    "printer_cls, expected",
    [
        (
            DefaultPrinter,
            "Results:\n"
            " 2 - org/repo1: https://www.github.com/org/repo1/search?utf8=✓&q=query\n"
            "\t- README.md\n"
            "\t- file.txt\n"
            " 1 - org/repo2: https://www.github.com/org/repo2/search?utf8=✓&q=query\n"
            "\t- file-2.json\n",
        ),
        (RepoListPrinter, "org/repo1\norg/repo2\n"),
        (
            JsonPrinter,
            '[{"full_name": "org/repo1", "html_url": "https://www.github.com/org/repo1", "fork": false, "owner": "org", "name": "repo1", "results": [{"path": "README.md", "name": "README.md", "size": 1000, "html_url": "https://www.github.com/org/repo1/blob/master/README.md"}, {"path": "file.txt", "name": "file.txt", "size": 1000, "html_url": "https://www.github.com/org/repo1/blob/master/file.txt"}]}, {"full_name": "org/repo2", "html_url": "https://www.github.com/org/repo2", "fork": false, "owner": "org", "name": "repo2", "results": [{"path": "file-2.json", "name": "file-2.json", "size": 1000, "html_url": "https://www.github.com/org/repo2/blob/master/file-2.json"}]}]',  # noqa: E501
        ),
        (
            YamlPrinter,
            "- fork: false\n"
            "  full_name: org/repo1\n"
            "  html_url: https://www.github.com/org/repo1\n"
            "  name: repo1\n"
            "  owner: org\n"
            "  results:\n"
            "  - html_url: https://www.github.com/org/repo1/blob/master/README.md\n"
            "    name: README.md\n"
            "    path: README.md\n"
            "    size: 1000\n"
            "  - html_url: https://www.github.com/org/repo1/blob/master/file.txt\n"
            "    name: file.txt\n"
            "    path: file.txt\n"
            "    size: 1000\n"
            "- fork: false\n"
            "  full_name: org/repo2\n"
            "  html_url: https://www.github.com/org/repo2\n"
            "  name: repo2\n"
            "  owner: org\n"
            "  results:\n"
            "  - html_url: https://www.github.com/org/repo2/blob/master/file-2.json\n"
            "    name: file-2.json\n"
            "    path: file-2.json\n"
            "    size: 1000\n",
        ),
    ],
)
def test_print_results_multiple_repos(printer_cls, expected):
    stream = StringIO()
    printer_cls(stream).print(
        ["query"],
        [
            build_mock_content_file("org/repo1", "README.md"),
            build_mock_content_file("org/repo1", "file.txt"),
            build_mock_content_file("org/repo2", "file-2.json"),
        ],
    )

    assert stream.getvalue() == expected


def test_run_with_qualifiers():
    stream = StringIO()
    DefaultPrinter(stream).print(
        ["query", "org:foo", "filename:bar"],
        [
            build_mock_content_file("org/repo1", "README.md"),
            build_mock_content_file("org/repo1", "file.txt"),
        ],
    )

    assert stream.getvalue() == (
        "Results:\n"
        " 2 - org/repo1: https://www.github.com/org/repo1/search?utf8=✓&q=query%20filename%3Abar\n"
        "\t- README.md\n"
        "\t- file.txt\n"
    )



from typing import Any, Dict

from github import Github

RESULTS_PER_PAGE = 100  # this is the max - see https://docs.github.com/en/rest/reference/search#search-code--parameters


def build_client(token: str, base_url: str | None = None) -> Github:
    client_params: Dict[str, Any] = {"per_page": RESULTS_PER_PAGE, "login_or_token": token}
    if base_url:
        client_params["base_url"] = base_url
    return Github(**client_params)



from unittest.mock import patch

import pytest

from ghsearch.client import build_client


@pytest.fixture
def mock_github():
    with patch("ghsearch.client.Github", autospec=True) as mock:
        yield mock


def test_build_client_returns_github_instance(mock_github):
    assert build_client("foo-token") == mock_github.return_value


def test_build_client_passes_expected_parameters(mock_github):
    build_client("foo-token", "https://github.example.org/api/v3")

    mock_github.assert_called_once_with(
        login_or_token="foo-token",
        base_url="https://github.example.org/api/v3",
        per_page=100,
    )


def test_build_client_passes_expected_parameters_defaults_github(mock_github):
    build_client("foo-token")

    mock_github.assert_called_once_with(
        login_or_token="foo-token",
        per_page=100,
    )



import os

import click
import click_config_file

from ghsearch.main import run
from ghsearch.output import Printer, printer_factory, printers_list


def _printer(ctx: click.Context, param: click.Parameter, value: str) -> Printer:
    try:
        force_repo_printer = bool(ctx.params.get("repos_with_matches"))
        return printer_factory(value, click.get_text_stream("stdout"), force_repo_printer)
    except KeyError:
        raise click.BadParameter(f"Must be one of: {', '.join(printers_list())}", ctx=ctx, param=param)


def _create_none_value_validator(message):
    def _validator(ctx, _, value):
        if value is None:
            raise click.UsageError(message, ctx=ctx)
        return value

    return _validator


_cmd_name = "gh-search"
_config_file = os.path.join(click.get_app_dir(_cmd_name), "config")


@click.command(
    help="QUERY must contain at least one search term, but may also contain search qualifiers"
    " (https://docs.github.com/en/github/searching-for-information-on-github/searching-code)",
    context_settings={"max_content_width": 120},
)
@click.argument("QUERY", nargs=-1, required=True)
@click.option(
    "--github-token",
    envvar="GITHUB_TOKEN",
    help="GitHub Auth Token. Will fall back on GITHUB_TOKEN envvar.",
    callback=_create_none_value_validator("GitHub token must be set via --github-token option or GITHUB_TOKEN envvar."),
)
@click.option(
    "--github-api-url",
    envvar="GITHUB_API_URL",
    help="Override default GitHub API URL. Can also specify via GITHUB_API_URL envvar.",
)
@click.option("-p", "--path-filter", help="Exclude results whose path (or part of path) does not match this.")
@click.option("-c", "--content-filter", help="Exclude results whose content does not match this.")
@click.option("-e", "--regex-content-filter", help="Exclude results whose content does not match this regex.")
@click.option("-a", "--include-archived", help="Include results from archived repos.", default=False, is_flag=True)
@click.option(
    "-l",
    "--repos-with-matches",
    help="Only the names of repos are printed. Equivalent to --output=repo-list",
    default=False,
    is_flag=True,
)
@click.option(
    "-o", "--output", help=f"Output style; one of: {', '.join(printers_list())}", callback=_printer, default="default"
)
@click.option("-v", "--verbose", help="Verbose output.", default=False, is_flag=True)
@click.version_option(package_name="gh-search")
@click_config_file.configuration_option(
    cmd_name="gh-search", help=f"Config file (default: {_config_file})", expose_value=True
)
def cli(
    query,
    output,
    include_archived,
    verbose,
    config,
    github_token,
    github_api_url=None,
    path_filter=None,
    content_filter=None,
    regex_content_filter=None,
    **_,
):
    if verbose:
        click.echo(f"Reading defaults from {config}")
    run(
        query=query,
        github_token=github_token,
        printer=output,
        github_api_url=github_api_url,
        path_filter=path_filter,
        content_filter=content_filter,
        regex_content_filter=regex_content_filter,
        include_archived=include_archived,
        verbose=verbose,
    )


if __name__ == "__main__":
    cli()



import re

from github.ContentFile import ContentFile
from github.GithubException import GithubException


class FilterException(BaseException):
    def __init__(self, filter, message):
        super().__init__(message)
        self.filter = filter


class Filter:
    """This filter uses the core api"""

    uses_core_api = True

    def __call__(self, result: ContentFile) -> bool:
        raise NotImplementedError


class DecodedContentFilter(Filter):
    def __call__(self, result: ContentFile) -> bool:
        try:
            content = result.decoded_content.decode("utf-8")
            return self.matches_content(content)
        except GithubException as e:
            message = f"Error reading content from {result.repository.full_name}/{result.path}: {e.data['message']}"
            raise FilterException(self, message) from e

    def matches_content(self, content: str) -> bool:
        raise NotImplementedError


class ContentFilter(DecodedContentFilter):
    def __init__(self, content_filter: str):
        self.content_filter = content_filter

    def matches_content(self, content: str) -> bool:
        return self.content_filter in content


class RegexContentFilter(DecodedContentFilter):
    def __init__(self, content_filter: str):
        try:
            self.content_filter_pattern = re.compile(content_filter)
        except re.error as e:
            message = f"Failed to compile regular expression from '{content_filter}': {e}"
            raise FilterException(self, message) from e

    def matches_content(self, content: str) -> bool:
        return bool(self.content_filter_pattern.search(content))


class NotArchivedFilter(Filter):
    def __init__(self):
        self.cache = {}

    def __call__(self, result: ContentFile) -> bool:
        if result.repository.full_name not in self.cache:
            self.cache[result.repository.full_name] = not result.repository.archived
        return self.cache[result.repository.full_name]


class PathFilter(Filter):
    def __init__(self, path_filter: str):
        self.uses_core_api = False
        self.path_filter = path_filter

    def __call__(self, result: ContentFile) -> bool:
        return self.path_filter in result.path



from types import SimpleNamespace
from unittest.mock import Mock

from github.ContentFile import ContentFile


class MockPaginatedList:
    def __init__(self, *items, total_count=None):
        self.items = items
        self.totalCount = total_count if total_count else len(items)

    def __iter__(self):
        return iter(self.items)


class MockRateLimit:
    def __init__(self, core_remaining, core_limit, core_reset, search_remaining, search_limit, search_reset):
        self.core = SimpleNamespace(remaining=core_remaining, limit=core_limit, reset=core_reset)
        self.search = SimpleNamespace(remaining=search_remaining, limit=search_limit, reset=search_reset)


def build_mock_content_file(
    repo_full_name: str = "org/repo", path: str = "path", archived: bool = False, decoded_content: bytes = b""
):
    mock = Mock(spec=ContentFile)
    mock.repository.name = repo_full_name.split("/")[1]
    mock.repository.owner.login = repo_full_name.split("/")[0]
    mock.repository.full_name = repo_full_name
    mock.repository.archived = archived
    mock.repository.html_url = f"https://www.github.com/{repo_full_name}"
    mock.repository.fork = False
    mock.path = path
    mock.decoded_content = decoded_content
    mock.name = path.split("/").pop()
    mock.size = 1000
    mock.html_url = f"https://www.github.com/{repo_full_name}/blob/master/{path}"
    return mock



from unittest.mock import Mock, call, patch

import click
import pytest
from github import BadCredentialsException, Github, GithubException

from ghsearch.main import run
from ghsearch.output import Printer

from . import MockPaginatedList, MockRateLimit, build_mock_content_file


@pytest.fixture(autouse=True)
def mock_progress_printer():
    with patch("ghsearch.gh_search.ProgressPrinter") as mock:
        yield mock


@pytest.fixture
def mock_content_file_repo1_readme():
    return build_mock_content_file("org/repo1", "README.md", decoded_content=b"special content")


@pytest.fixture
def mock_content_file_repo1_file():
    return build_mock_content_file("org/repo1", "file.txt")


@pytest.fixture
def mock_content_file_repo2_src_other_archived():
    return build_mock_content_file("org/repo2", "src/other.py", archived=True)


@pytest.fixture
def mock_github(
    mock_content_file_repo1_readme, mock_content_file_repo1_file, mock_content_file_repo2_src_other_archived
):
    mock = Mock(spec=Github)
    mock.search_code.return_value = MockPaginatedList(
        mock_content_file_repo1_readme,
        mock_content_file_repo1_file,
        mock_content_file_repo2_src_other_archived,
    )
    mock.get_rate_limit.side_effect = [
        MockRateLimit(45, 50, "soon", 10, 10, "soon"),
        MockRateLimit(43, 50, "even sooner", 9, 10, "even sooner"),
    ]
    return mock


@pytest.fixture(autouse=True)
def mock_build_client(mock_github):
    with patch("ghsearch.main.build_client") as mock:
        mock.return_value = mock_github
        yield mock


@pytest.fixture()
def mock_printer():
    return Mock(spec=Printer)


def test_run_exclude_archived_by_default(
    assert_click_echo_calls, mock_printer, mock_content_file_repo1_readme, mock_content_file_repo1_file
):
    run(["query"], "token", mock_printer)
    mock_printer.print.assert_called_once_with(
        ["query"],
        [
            mock_content_file_repo1_readme,
            mock_content_file_repo1_file,
        ],
    )


def test_run_bad_credentials(assert_click_echo_calls, mock_github, mock_printer):
    mock_github.search_code.side_effect = BadCredentialsException(404, "No!")
    with pytest.raises(click.UsageError, match='Bad Credentials: 404 "No!"'):
        run(["query"], "bad-credentials", mock_printer)


def test_run_verbose(assert_click_echo_calls, mock_printer):
    run(["query"], "token", mock_printer, verbose=True)
    assert_click_echo_calls(
        call("Core rate limit: 45/50 (resets soon), Search rate limit: 10/10 (resets soon)"),
        call("Skipping result for org/repo2 via NotArchivedFilter"),
        call("Core rate limit: 43/50 (resets even sooner), Search rate limit: 9/10 (resets even sooner)"),
    )


def test_run_include_archived(
    assert_click_echo_calls,
    mock_printer,
    mock_content_file_repo1_readme,
    mock_content_file_repo1_file,
    mock_content_file_repo2_src_other_archived,
):
    run(["query"], "token", mock_printer, include_archived=True)
    mock_printer.print.assert_called_once_with(
        ["query"],
        [
            mock_content_file_repo1_readme,
            mock_content_file_repo1_file,
            mock_content_file_repo2_src_other_archived,
        ],
    )


def test_run_content_filter(assert_click_echo_calls, mock_printer, mock_content_file_repo1_readme):
    run(["query"], "token", mock_printer, content_filter="special content")
    mock_printer.print.assert_called_once_with(
        ["query"],
        [
            mock_content_file_repo1_readme,
        ],
    )


def test_run_regex_content_filter(assert_click_echo_calls, mock_printer, mock_content_file_repo1_readme):
    run(["query"], "token", mock_printer, regex_content_filter="special\\scontent")
    mock_printer.print.assert_called_once_with(
        ["query"],
        [
            mock_content_file_repo1_readme,
        ],
    )


def test_run_regex_content_filter_bad_regex(mock_printer):
    with pytest.raises(
        click.UsageError,
        match="Failed to compile regular expression from '\\[': unterminated character set at position 0",
    ):
        run(["query"], "token", mock_printer, regex_content_filter="[")


def test_run_path_filter(assert_click_echo_calls, mock_printer, mock_content_file_repo1_file):
    run(["query"], "token", mock_printer, path_filter="file.txt")
    mock_printer.print.assert_called_once_with(["query"], [mock_content_file_repo1_file])


def test_run_when_raises_github_exception_422(mock_github, mock_printer):
    mock_github.search_code.side_effect = GithubException(
        422, {"message": "Fail!", "errors": [{"message": "reason1"}, {"message": "reason2"}]}
    )

    with pytest.raises(click.UsageError, match="Fail! \\(GitHub Exception\\): reason1, reason2"):
        run(["query"], "token", mock_printer)


def test_run_when_raises_github_exception(mock_github, mock_printer):
    mock_github.search_code.side_effect = GithubException(400, "")

    with pytest.raises(GithubException):
        run(["query"], "token", mock_printer)



from unittest.mock import Mock, patch

import github
import pytest

from ghsearch.filters import FilterException
from ghsearch.gh_search import GHSearch

from . import MockPaginatedList, MockRateLimit, build_mock_content_file


@pytest.fixture
def mock_result_1():
    return build_mock_content_file("org/repo1", "1.txt")


@pytest.fixture
def mock_result_2():
    return build_mock_content_file("org/repo1", "2.txt")


@pytest.fixture
def mock_result_3():
    return build_mock_content_file("org/repo2", "3.txt")


@pytest.fixture
def mock_client(mock_result_1, mock_result_2, mock_result_3):
    mock = Mock(spec=github.Github)
    mock.search_code.return_value = MockPaginatedList(mock_result_1, mock_result_2, mock_result_3)
    mock.get_rate_limit.return_value = MockRateLimit(10, 10, "now", 10, 10, "now")
    return mock


@pytest.fixture(autouse=True)
def mock_click():
    with patch("ghsearch.gh_search.click") as mock:
        yield mock


@pytest.fixture(autouse=True)
def mock_progress_printer():
    with patch("ghsearch.gh_search.ProgressPrinter") as mock:
        yield mock


def test_get_filtered_results_calls_search_code_correctly(mock_client):
    ghsearch = GHSearch(mock_client, [])
    ghsearch.get_filtered_results(["name", "org:janeklb", "filename:setup.py"])

    mock_client.search_code.assert_called_once_with(query="name org:janeklb filename:setup.py")


def test_get_filtered_results_without_filters(mock_client, mock_result_1, mock_result_2, mock_result_3):
    ghsearch = GHSearch(mock_client, [])
    repos = ghsearch.get_filtered_results(["query", "org:bort"])

    assert repos == [mock_result_1, mock_result_2, mock_result_3]


def test_get_filtered_results_with_filters(mock_client, mock_result_1, mock_result_2, mock_result_3):
    ghsearch = GHSearch(mock_client, [Mock(side_effect=[True, False, True])])
    repos = ghsearch.get_filtered_results(["query", "org:bort"])

    assert repos == [mock_result_1, mock_result_3]


def test_get_filtered_results_with_exception_when_filtering(mock_client, mock_result_1, mock_progress_printer):
    mock_client.search_code.return_value = MockPaginatedList(mock_result_1)

    ghsearch = GHSearch(mock_client, [Mock(side_effect=FilterException(Mock(), "BOOO"))])

    repos = ghsearch.get_filtered_results(["query", "org:bort"])

    assert len(repos) == 0
    mock_progress_printer.return_value.__enter__.return_value.assert_any_call("BOOO", force=True)


def test_get_filtered_results_verbose(mock_client, mock_result_1, mock_result_2, mock_result_3, mock_click):
    ghsearch = GHSearch(
        mock_client, [Mock(side_effect=[True, True, False]), Mock(side_effect=[False, True, False])], verbose=True
    )

    repos = ghsearch.get_filtered_results(["query", "org:bort"])

    assert repos == [mock_result_2]
    mock_click.echo.assert_any_call("Skipping result for org/repo1 via Mock")
    mock_click.echo.assert_any_call("Skipping result for org/repo2 via Mock")


def test_get_filtered_results_near_limit(mock_client, mock_click):
    mock_client.get_rate_limit.return_value = MockRateLimit(1, 10, "sometime in the future", 10, 10, "now")
    mock_filter = Mock()
    mock_filter.uses_core_api = True

    ghsearch = GHSearch(mock_client, [mock_filter])
    ghsearch.get_filtered_results(["query", "org:bort"])

    mock_click.confirm.assert_called_once_with(
        """
Warning: you are at risk of going below 10% of your remaining core api rate limit.
Your search yielded 3 results, and gh-search may make up to 1 core api call(s) per result.

Your current core api usage is 1/10 (resets sometime in the future)

Do you want to continue?""".strip(),
        abort=True,
    )


def test_get_filtered_results_many_calls(mock_client, mock_click):
    mock_client.get_rate_limit.return_value = MockRateLimit(10000, 10000, "sometime in the future", 10, 10, "now")
    mock_client.search_code.return_value = MockPaginatedList(*[], total_count=257)
    mock_filter = Mock()
    mock_filter.uses_core_api = True

    ghsearch = GHSearch(mock_client, [mock_filter, mock_filter])
    ghsearch.get_filtered_results(["query", "org:bort"])

    mock_click.confirm.assert_called_once_with(
        """
Warning: you are about to potentially make more than 500 core api requests.
Your search yielded 257 results, and gh-search may make up to 2 core api call(s) per result.

Your current core api usage is 10000/10000 (resets sometime in the future)

Do you want to continue?""".strip(),
        abort=True,
    )


def test_get_filtered_results_rate_limiting_disabled(mock_client):
    mock_client.get_rate_limit.side_effect = github.GithubException(404, "Not Found")
    mock_filter = Mock()
    mock_filter.uses_core_api = True

    ghsearch = GHSearch(mock_client, [])
    ghsearch.get_filtered_results(["query", "org:bort"])

    # ensure get_rate_limit was called (and the side_effect above handled)
    mock_client.get_rate_limit.assert_called()



from typing import List

import click
from click import UsageError
from github.GithubException import BadCredentialsException, GithubException

from ghsearch.client import build_client
from ghsearch.filters import ContentFilter, Filter, FilterException, NotArchivedFilter, PathFilter, RegexContentFilter
from ghsearch.gh_search import GHSearch
from ghsearch.output import Printer


def _build_filters(
    path_filter: str | None = None,
    include_archived: bool = True,
    content_filter: str | None = None,
    regex_content_filter: str | None = None,
) -> List[Filter]:
    filters: List[Filter] = []
    if path_filter:
        filters.append(PathFilter(path_filter))
    if not include_archived:
        filters.append(NotArchivedFilter())
    if content_filter:
        filters.append(ContentFilter(content_filter))
    if regex_content_filter:
        filters.append(RegexContentFilter(regex_content_filter))
    return filters


def run(
    query: List[str],
    github_token: str,
    printer: Printer,
    github_api_url: str | None = None,
    path_filter: str | None = None,
    content_filter: str | None = None,
    regex_content_filter: str | None = None,
    include_archived: bool = False,
    verbose: bool = False,
) -> None:
    client = build_client(github_token, github_api_url)

    try:
        filters = _build_filters(path_filter, include_archived, content_filter, regex_content_filter)
    except FilterException as ex:
        raise UsageError(str(ex), click.get_current_context(silent=True))

    try:
        gh_search = GHSearch(client, filters, verbose)
        results = gh_search.get_filtered_results(query)

        printer.print(query, results)

    except BadCredentialsException as ex:
        raise UsageError(f"Bad Credentials: {ex}", click.get_current_context(silent=True))
    except GithubException as ex:
        if ex.status == 422 and isinstance(ex.data, dict):
            message = ex.data["message"]
            errors = ", ".join(err["message"] for err in ex.data.get("errors", []) if isinstance(err, dict))
            raise UsageError(f"{message} (GitHub Exception): {errors}", click.get_current_context(silent=True))
        raise ex
